"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FileDecorationsService = void 0;
const tslib_1 = require("tslib");
const di_1 = require("@opensumi/di");
const ide_core_common_1 = require("@opensumi/ide-core-common");
const { isFalsyOrWhitespace } = ide_core_common_1.strings;
const { asArray } = ide_core_common_1.arrays;
class FileDecorationChangeEvent {
    constructor() {
        this._data = ide_core_common_1.TernarySearchTree.forPaths();
    }
    affectsResource(uri) {
        return this._data.get(uri.toString()) || this._data.findSuperstr(uri.toString()) !== undefined;
    }
    static debouncer(last, current) {
        if (!last) {
            last = new FileDecorationChangeEvent();
        }
        if (Array.isArray(current)) {
            // many
            for (const uri of current) {
                last._data.set(uri.toString(), true);
            }
        }
        else {
            // one
            last._data.set(current.toString(), true);
        }
        return last;
    }
}
class DecorationDataRequest {
    constructor(source, thenable) {
        this.source = source;
        this.thenable = thenable;
    }
}
class DecorationProviderWrapper {
    constructor(_provider, _uriEmitter, _flushEmitter) {
        this._provider = _provider;
        this._uriEmitter = _uriEmitter;
        this._flushEmitter = _flushEmitter;
        this.data = ide_core_common_1.TernarySearchTree.forPaths();
        this._dispoable = this._provider.onDidChange((uris) => {
            if (!uris) {
                // flush event -> drop all data, can affect everything
                this.data.clear();
                this._flushEmitter.fire({
                    affectsResource() {
                        return true;
                    },
                });
            }
            else {
                // selective changes -> drop for resource, fetch again, send event
                // perf: the map stores thenables, decorations, or `null`-markers.
                // we make us of that and ignore all uris in which we have never
                // been interested.
                for (const uri of uris) {
                    this._fetchData(uri);
                }
            }
        });
    }
    dispose() {
        this._dispoable.dispose();
        this.data.clear();
    }
    knowsAbout(uri) {
        return Boolean(this.data.get(uri.toString())) || Boolean(this.data.findSuperstr(uri.toString()));
    }
    getOrRetrieve(uri, includeChildren, callback) {
        const key = uri.toString();
        let item = this.data.get(key);
        if (item === undefined) {
            // unknown -> trigger request
            item = this._fetchData(uri);
        }
        if (item && !(item instanceof DecorationDataRequest)) {
            // found something (which isn't pending anymore)
            callback(item, false);
        }
        if (includeChildren) {
            // (resolved) children
            const iter = this.data.findSuperstr(key);
            if (iter) {
                for (let item = iter.next(); !item.done; item = iter.next()) {
                    if (item.value && !(item.value instanceof DecorationDataRequest)) {
                        callback(item.value, true);
                    }
                }
            }
        }
    }
    _fetchData(uri) {
        // check for pending request and cancel it
        const pendingRequest = this.data.get(uri.toString());
        if (pendingRequest instanceof DecorationDataRequest) {
            pendingRequest.source.cancel();
            this.data.delete(uri.toString());
        }
        const source = new ide_core_common_1.CancellationTokenSource();
        const dataOrThenable = this._provider.provideDecorations(uri, source.token);
        if (!(0, ide_core_common_1.isThenable)(dataOrThenable)) {
            // sync -> we have a result now
            return this._keepItem(uri, dataOrThenable);
        }
        else {
            // async -> we have a result soon
            const request = new DecorationDataRequest(source, Promise.resolve(dataOrThenable)
                .then((data) => {
                if (this.data.get(uri.toString()) === request) {
                    this._keepItem(uri, data);
                }
            })
                .catch((err) => {
                if (!(0, ide_core_common_1.isPromiseCanceledError)(err) && this.data.get(uri.toString()) === request) {
                    this.data.delete(uri.toString());
                }
            }));
            this.data.set(uri.toString(), request);
            return null;
        }
    }
    _keepItem(uri, data) {
        const deco = data ? data : null;
        const old = this.data.set(uri.toString(), deco);
        if (deco || old) {
            // only fire event when something changed
            this._uriEmitter.fire(uri);
        }
        return deco;
    }
}
// 将 IDecorationData 转成 `${color}/${letter}` 形式
function keyOfDecorationRule(data) {
    const list = asArray(data);
    return list.map(({ color, letter }) => `${color}/${letter}`).join(',');
}
function getDecorationRule(data) {
    const list = asArray(data);
    if (!list.length) {
        // 前置拦截(目前代码逻辑是不会进来的)
        /* istanbul ignore next */
        return;
    }
    return {
        // 获取 color/letter 生成的 唯一键 key
        key: keyOfDecorationRule(data),
        // label
        color: list[0].color,
        // badge
        badge: list
            .filter((d) => !isFalsyOrWhitespace(d.letter))
            .map((d) => d.letter)
            .join(','),
        tooltip: list
            .filter((d) => !isFalsyOrWhitespace(d.tooltip))
            .map((d) => d.tooltip)
            .join(' • '),
    };
}
let FileDecorationsService = class FileDecorationsService extends ide_core_common_1.Disposable {
    constructor() {
        super(...arguments);
        this.logger = (0, ide_core_common_1.getDebugLogger)();
        this._data = new ide_core_common_1.LinkedList();
        this._onDidChangeDecorationsDelayed = new ide_core_common_1.Emitter();
        this._onDidChangeDecorations = new ide_core_common_1.Emitter();
        this.onDidChangeDecorations = ide_core_common_1.Event.any(this._onDidChangeDecorations.event, ide_core_common_1.Event.debounce(this._onDidChangeDecorationsDelayed.event, FileDecorationChangeEvent.debouncer, // todo: remove it
        undefined, undefined, 500));
    }
    dispose() {
        (0, ide_core_common_1.dispose)(this._onDidChangeDecorations);
        (0, ide_core_common_1.dispose)(this._onDidChangeDecorationsDelayed);
    }
    registerDecorationsProvider(provider) {
        this.logger.log('DecorationService#registerDecorationsProvider', provider);
        const wrapper = new DecorationProviderWrapper(provider, this._onDidChangeDecorationsDelayed, this._onDidChangeDecorations);
        const remove = this._data.push(wrapper);
        this._onDidChangeDecorations.fire({
            // everything might have changed
            affectsResource() {
                return true;
            },
        });
        return (0, ide_core_common_1.toDisposable)(() => {
            // fire event that says 'yes' for any resource
            // known to this provider. then dispose and remove it.
            remove();
            this._onDidChangeDecorations.fire({ affectsResource: (uri) => wrapper.knowsAbout(uri) });
            wrapper.dispose();
        });
    }
    getDecoration(uri, includeChildren) {
        const data = [];
        let containsChildren = false;
        for (let iter = this._data.iterator(), next = iter.next(); !next.done; next = iter.next()) {
            next.value.getOrRetrieve(uri, includeChildren, (deco, isChild) => {
                if (!isChild || deco.bubble) {
                    data.push(deco);
                    containsChildren = isChild || containsChildren;
                }
            });
        }
        if (data.length === 0) {
            return undefined;
        }
        else {
            const result = this.asDecoration(data, containsChildren);
            return result;
        }
    }
    asDecoration(data, containsChildren) {
        // sort by weight
        data.sort((a, b) => (b.weight || 0) - (a.weight || 0));
        const rule = getDecorationRule(data);
        if (rule && containsChildren) {
            // 目录下文件修改过则仅显示一个圆点并统一 tooltip
            // show items from its children only
            rule.badge = '•';
            rule.tooltip = (0, ide_core_common_1.localize)('bubbleTitle', 'Contains emphasized items');
        }
        return rule;
    }
};
FileDecorationsService = tslib_1.__decorate([
    (0, di_1.Injectable)()
], FileDecorationsService);
exports.FileDecorationsService = FileDecorationsService;
//# sourceMappingURL=decorationsService.js.map